/* Written by Tibor 'Igor2' Palinkas. This file is in the public domain. */

/* Derived from: 01_main_sc_stdin.c
   Compile: cc -o 05_c_func 05_c_func.c -lm
   Execute: ./05_c_func < 05.fawk
   Features:
    - loads a script from a stdin
    - provide C function "hello_str()" that can be called from the script
    - execute the main part
    - no "include" support
*/

/*** libfawk configuration ***/
typedef double fawk_num_t;
typedef long fawk_refco_t;
#define FAWK_API static

/* uncomment this to disable fawk_print() -> cuts off stdio dependency */
/*#define FAWK_DISABLE_FAWK_PRINT*/

/* include the single-c, all language variant */
#include "../../../single_c/libfawk_sc_all/libfawk_sc_all.c"

/*** glue ***/
FAWK_API void libfawk_error(fawk_ctx_t *ctx, const char *str, const char *loc_fn, long loc_line, long loc_col)
{
	fprintf(stderr, "Error: %s\n", str);
}

static int getch1(fawk_ctx_t *ctx, fawk_src_t *src)
{
	return getchar();
}

/*** script callable function */
static void hello_str(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	fawk_cell_t *times = FAWK_CFUNC_ARG(0);
	char tmp[256];

	if (argc > 1) {
		fprintf(stderr, "hello_str: too many arguments\n");
		return; /* retval is initialized to NIL on the caller side */
	}

	if (argc < 1) {
		fprintf(stderr, "hello_str: not enough arguments\n");
		return; /* retval is initialized to NIL on the caller side */
	}

	fawk_cast_to_num(ctx, times); /* convert the result to num, just in case the fawk function returned a string */

	sprintf(tmp, "hello world %d times!", (int)times->data.num);
	retval->type = FAWK_STR;
	retval->data.str = fawk_str_new_from_literal(ctx, tmp, -1);
}

int main(int argc, char *argv[])
{
	FILE *fsc;
	fawk_ctx_t ctx;

	/* set up the script context */
	fawk_init(&ctx);
	ctx.parser.get_char = getch1;

	/* register the new script function */
	fawk_symtab_regcfunc(&ctx, "hello_str", hello_str);

	/* parse; replace fawk_parse_fawk() with fawk_parse_fpas() or
	   fawk_parse_fbas() for the other languages */
	if (fawk_parse_fawk(&ctx) != 0) {
		fawk_uninit(&ctx);
		return 1;
	}

	/* call main() - call1 prepares the stack */
	if (fawk_call1(&ctx, "main") == 0) {
		fawk_cell_t res;
		fawk_execret_t er;

		/* arguments could be pushed on the stack here */

		if (fawk_call2(&ctx, 0) == 0) { /* call2: perform call, with 0 arguments on stack */
			er = fawk_execute(&ctx, -1); /* execute the VM; will return once the function has finished */
			if (er == FAWK_ER_FIN) { /* read and free the return value of script's main() */
				fawk_pop(&ctx, &res);
				printf("\nResult: ");
				fawk_dump_cell(&res, 1);
				printf("\n");
				fawk_cell_free(&ctx, &res);
			}
		}
		else
			fprintf(stderr, "Failed to call main() [#2]\n");
	}
	else
		fprintf(stderr, "Failed to call main() [#1]\n");

	fawk_uninit(&ctx);

	return 0;
}

